#!/usr/bin/env python3

import numpy as np
from numpy.core.umath_tests import inner1d
from typing import Tuple
from Tools import hash_2d, hash_3d, hash_nd



def F(d):
    return (np.sqrt(d+1) - 1) / d


def G(d):
    return (1 - 1/np.sqrt(d+1))/d


def scew_coords(coords):
    d = coords.shape[1]
    n = coords.shape[0]
    f = F(d)

    coords_sum = np.sum(coords, axis=1).reshape((n, 1))

    return coords + coords_sum * f


def un_scew_coords(coords):
    d = coords.shape[1]
    n = coords.shape[0]
    g = G(d)

    coords_sum = np.sum(coords, axis=1).reshape((n, 1))

    return coords - coords_sum * g


class SimplexGenerator(object):

    def __init__(self,
                 world_seed: int = 0,
                 scewed_chunk_size: int = 8):
        self.world_seed = world_seed
        self.scewed_chunk_size = scewed_chunk_size
        self.cached_hashed_chunks = {}

    def _is_cached(self, scewed_coord):
        return self.cached_hashed_chunks.__contains__(scewed_coord)

    def _create_hashed_vector(self, scewed_coord):
        np.random.seed(hash_nd(np.array(scewed_coord +
                                        self.world_seed, dtype=int)) %
                       np.uint64(2**32))

        v = np.array([np.random.normal() for i in range(len(scewed_coord))])
        mag = np.linalg.norm(v)
        return v / mag

    def _get_hashed_vector(self, scewed_coord):
        key_tuple = tuple(scewed_coord)
        try:
            return self.cached_hashed_chunks[key_tuple]
        except KeyError:
            val = self._create_hashed_vector(  # noqa
                scewed_coord)
            self.cached_hashed_chunks[key_tuple] = val
            return val

    def _get_scewed_corners(self, coords):

        # note: a lot of numpy magic is going on here.
        # probably this would be a lot of more code
        # in every other programming language :)

        d = coords.shape[1]
        n = coords.shape[0]

        corners = np.zeros(shape=(n, d+1, d), dtype=float)

        scewed_coords = scew_coords(coords)

        # get integer part (base corner)
        truncated = np.floor(scewed_coords)

        # get remaining part
        local_coords = scewed_coords - truncated

        # sort remaining part
        local_order = np.argsort(local_coords, axis=-1)

        base_vectors = np.identity(d)

        # get base vectors sorted for every coordinate by local orders:
        sorted_base_vectors = base_vectors[local_order]

        corners[:, 0, :] = truncated
        for i in range(1, d+1):
            # -i becease we want decreasing order!
            corners[:, i, :] = corners[:, i-1, :] + \
                sorted_base_vectors[:, :, -i]

        # we're done
        return corners

    def get_dense_values(self, points, r_sqrd=0.5):
        d = points.shape[1]
        n = points.shape[0]

        scewed_corners = self._get_scewed_corners(points)

        # get min and max scewd values:
        min_scewed = np.min(scewed_corners, axis=0)
        max_scewed = np.max(scewed_corners, axis=0)

        min_scewed = np.min(np.array(min_scewed, dtype=int), axis=0)
        max_scewed = np.max(np.array(max_scewed, dtype=int), axis=0)

        grid_shape = tuple(np.array(max_scewed - min_scewed,
                                    dtype=int) + np.ones(d, dtype=int))

        local_grid = np.zeros(tuple(list(grid_shape)+[d]))

        # generate grid in [min, max] range:
        for index in np.ndindex(grid_shape):
            local_grid[index] = self._get_hashed_vector(min_scewed + index)

        result = np.zeros(shape=(n, d))

        for i in range(d+1):
            # unscew corners
            uc = un_scew_coords(scewed_corners[:, i, :])
            dists = (uc - points)
            sqrd_dists = np.sum(dists**2, axis=-1)

            corner_indices = np.zeros(shape=(d, n), dtype=int)

            for j in range(d):
                # TODO: maybe rollaxis is better
                corner_indices[j, :] = scewed_corners[:,
                                                      i, j] - min_scewed[j]
            
            assert not np.min(corner_indices) < 0

            # getting gradients:
            g = local_grid[tuple(corner_indices)]

            # implement the formula given above:
            s = (np.maximum(0, r_sqrd - sqrd_dists))**4 * inner1d(dists, g)

            result += (s * g.T).T

        return result
