#!/usr/bin/env python3

import numpy as np
import Chunk
import random
from SimplexGenerator import SimplexGenerator
from threading import Lock

from typing import Tuple


class WorldManager(object):

    def __init__(self, world_seed: int, world_scale: int, biome_scale: int = 32):

        self.world_seed = world_seed

        self.simplex_generator_2d = SimplexGenerator()
        self.simplex_generator_3d = SimplexGenerator()

        # dictionary of dictionaries of dictionaries --> 3D hasmap
        # (assume that's more efficient than using tuples as keys,
        # but future me should do further investigation into that [TODO!])
        self.chunks = {}

        # will be a list with locked chunks
        self.chunk_locks = {}

        self.world_scale = world_scale
        self.biome_scale = biome_scale

        # create initial chunk
        # self.createEmptyChunk(0, 0, 0)

        # create lock object
        self.world_lock = Lock()

    def acquireChunk(self, chunk_x: int, chunk_y: int, chunk_z: int) -> bool:
        assert self.isChunkKnown(chunk_x, chunk_y, chunk_z)
        return self.chunk_locks[chunk_x][chunk_y][chunk_z].acquire()

    def releaseChunk(self, chunk_x: int, chunk_y: int, chunk_z: int) -> bool:
        assert self.isChunkKnown(chunk_x, chunk_y, chunk_z)
        lock = self.chunk_locks[chunk_x][chunk_y][chunk_z]
        assert lock.locked()
        return lock.release()

    def isChunkKnown(self, chunk_x: int, chunk_y: int, chunk_z: int) -> bool:
        # TODO: this should be done more efficient!
        self.world_lock.acquire()
        is_known = (chunk_x in self.chunks and
                    chunk_y in self.chunks[chunk_x] and
                    chunk_z in self.chunks[chunk_x][chunk_y])
        self.world_lock.release()
        return is_known

    def createEmptyChunk(self,
                         chunk_x: int,
                         chunk_y: int,
                         chunk_z: int) -> None:

        assert not self.isChunkKnown(chunk_x, chunk_y, chunk_z)

        self.world_lock.acquire()

        if chunk_x not in self.chunks:
            self.chunks[chunk_x] = {}
            self.chunk_locks[chunk_x] = {}

        if chunk_y not in self.chunks[chunk_x]:
            self.chunks[chunk_x][chunk_y] = {}
            self.chunk_locks[chunk_x][chunk_y] = {}

        self.chunks[chunk_x][chunk_y][chunk_z] = Chunk.Chunk(
            chunk_x, chunk_y, chunk_z)

        self.chunk_locks[chunk_x][chunk_y][chunk_z] = Lock()

        self.world_lock.release()

    def _getDensesForChunk(self, chunk_coord, simplex_generator, world_scale):
        n = len(chunk_coord)
        chunk_shape = tuple([Chunk.Chunk.CHUNK_SIDELENGTH] * n)

        # NOTE: trick for fast point generation found on:
        # https://github.com/numpy/numpy/issues/1234
        points = np.indices(chunk_shape).reshape(len(chunk_shape), -1).T
        points = np.array(points, dtype=float) / \
            (Chunk.Chunk.CHUNK_SIDELENGTH * world_scale)

        # adding chunk offset to points:
        points += np.array(chunk_shape, dtype=float) * chunk_coord / \
            (Chunk.Chunk.CHUNK_SIDELENGTH * world_scale)

        denses = simplex_generator.get_dense_values(points)
        magnitudes = np.linalg.norm(denses, axis=-1)

        denses = denses.reshape(
            tuple(list(chunk_shape) + [len(chunk_shape)]))
        magnitudes = magnitudes.reshape(chunk_shape)

        return denses, magnitudes

    def applySimplexToChunk(self,
                            chunk_x: int,
                            chunk_y: int,
                            chunk_z: int) -> None:

        # biome map generation:
        biome_denses, biome_magnitudes = self._getDensesForChunk(np.array(
            [chunk_x, chunk_z], dtype=int), self.simplex_generator_2d, self.biome_scale)

        biome_magnitudes *= 200 * Chunk.Chunk.CHUNK_SIDELENGTH

        # 2D heightmap generation

        denses, magnitudes = self._getDensesForChunk(np.array([chunk_x, chunk_z],
                                                              dtype=int),
                                                     self.simplex_generator_2d,
                                                     self.world_scale)

        magnitudes *= 50 * Chunk.Chunk.CHUNK_SIDELENGTH

        self.acquireChunk(chunk_x, chunk_y, chunk_z)

        c = self.chunks[chunk_x][chunk_y][chunk_z]

        for y in range(Chunk.Chunk.CHUNK_SIDELENGTH):
            c.block_data[:, y, :] = (magnitudes[:, :] + biome_magnitudes[:, :] > y +
                                     chunk_y*Chunk.Chunk.CHUNK_SIDELENGTH)

        self.releaseChunk(chunk_x, chunk_y, chunk_z)

        # 3D complex generation (e.g. for caves)

        '''
        denses, magnitudes = self._getDensesForChunk(np.array([chunk_x, chunk_y, chunk_z],
                                                             dtype=int),
                                                    self.simplex_generator_3d,
                                                    self.world_scale)
            
        magnitudes *= 47 # 47 ~ 1 / max value
        
        c.block_data = np.array(np.multiply(magnitudes < 0.3, c.block_data), dtype=int)
        '''

    def getChunkAsJson(self, chunk_x: int, chunk_y: int, chunk_z: int) -> str:
        assert self.isChunkKnown(chunk_x, chunk_y, chunk_z)
        self.acquireChunk(0, 0, 0)
        json = repr(self.chunks[chunk_x][chunk_y][chunk_z])
        self.releaseChunk(0, 0, 0)
        return json
